%{
#include "hoc.h"
int lineno = 0;
%}
%union {
    double val;
    symbol* sym;
}
%token <val>    NUMBER
%token <sym>    VAR BLTIN UNDEF
%type  <val>    expr asgn
%right '='
%left  '+' '-'
%left  '*' '/'
%left  UNARYMINUS UNARYPLUS
%right '^'
%%
list:             /* nothing */
    | list '\n'
    | list asgn '\n'
    | list expr  '\n' { printf( "%d: \t%.8g\n", lineno, $2 ); }
    | list error '\n' { yyerrok; }
    ;
asgn:   VAR '=' expr { $$=$1->u.val = $3 ; $1->type = VAR; }
    ;
expr: NUMBER
    | VAR   { if ($1->type == UNDEF )
                execerror("undefined variable", $1->name );
                $$ = $1->u.val;
            }
    | asgn
    | BLTIN '(' expr ')' { $$ = (*($1->u.ptr))($3); }
    | expr '+' expr { $$ = $1 + $3; }
    | expr '-' expr { $$ = $1 - $3; }
    | expr '*' expr { $$ = $1 * $3; }
    | expr '/' expr { $$ = $1 / $3; }
    | expr '^' expr { $$ = pow($1, $3) ; }
    | '(' expr ')'  { $$ = $2 ; }
    | '-' expr %prec UNARYMINUS  { $$ = -$2; }
    | '+' expr %prec UNARYPLUS   { $$ =  $2; }
    ;
%%

int yylex( void )
{
    int c, c1, c2 ;

    while (( c = getchar()) == ' ' || c == '\t' )
        ;
    if ( c == EOF )
        return 0;
    if ( c == 'b' ) {
        if ((c1 = getchar()) == 'y' ) {
            if ((c2 = getchar()) == 'e' ) {
                return 0;
            }
            else ungetc( c2, stdin );
        }
        else ungetc( c1, stdin );
    }
    if ( c == '.' || isdigit(c)) {
        ungetc( c, stdin );
        scanf( "%lf", &yylval.val );
        return NUMBER;
    }
    if (isalpha(c)) {
        symbol *s;
        char sbuf[100], *p = sbuf;
        do {
            *p++ = c;
        } while ((c=getchar()) != EOF && isalnum(c));
        ungetc( c, stdin );
        *p = '\0';
        if (( s= lookup(sbuf)) == 0 )
            s = install(sbuf, UNDEF, 0.0 );
        yylval.sym = s;
        return s->type == UNDEF ? VAR : s->type;
    }
            
    if ( c == '\n' )
        lineno++;
    return c;
}

